import asyncio
import inspect
import json
import logging
from asyncio import CancelledError, Queue
from functools import partial
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    NoReturn,
    Optional,
    Text,
    Union,
)

import structlog
from sanic import Blueprint, response
from sanic.request import Request
from sanic.response import BaseHTTPResponse, HTTPResponse, ResponseStream

import rasa.utils.endpoints
from rasa.core.channels.channel import (
    CollectingOutputChannel,
    InputChannel,
    UserMessage,
)

logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()


class RestInput(InputChannel):
    """A custom http input channel.

    This implementation is the basis for a custom implementation of a chat
    frontend. You can customize this to send messages to Rasa and
    retrieve responses from the assistant.
    """

    @classmethod
    def name(cls) -> Text:
        return "rest"

    @staticmethod
    async def on_message_wrapper(
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
        text: Text,
        queue: Queue,
        sender_id: Text,
        input_channel: Text,
        metadata: Optional[Dict[Text, Any]],
    ) -> None:
        collector = QueueOutputChannel(queue)

        message = UserMessage(
            text, collector, sender_id, input_channel=input_channel, metadata=metadata
        )
        await on_new_message(message)

        await queue.put("DONE")

    async def _extract_sender(self, req: Request) -> Optional[Text]:
        return req.json.get("sender", None)

    # noinspection PyMethodMayBeStatic
    def _extract_message(self, req: Request) -> Optional[Text]:
        return req.json.get("message", None)

    def _extract_input_channel(self, req: Request) -> Text:
        return req.json.get("input_channel") or self.name()

    def get_metadata(self, request: Request) -> Optional[Dict[Text, Any]]:
        """Extracts additional information from the incoming request.

         Implementing this function is not required. However, it can be used to extract
         metadata from the request. The return value is passed on to the
         ``UserMessage`` object and stored in the conversation tracker.

        Args:
            request: incoming request with the message of the user

        Returns:
            Metadata which was extracted from the request.
        """
        return request.json.get("metadata", None)

    async def stream_response(
        self,
        on_new_message: Callable[[UserMessage], Awaitable[None]],
        text: Text,
        sender_id: Text,
        input_channel: Text,
        metadata: Optional[Dict[Text, Any]],
        resp: ResponseStream,
    ) -> None:
        """Streams response to the client.

         If the stream option is enabled, this method will be called to
         stream the response to the client

        Args:
            on_new_message: sanic event
            text: message text
            sender_id: message sender_id
            input_channel: input channel name
            metadata: optional metadata sent with the message
            resp: response stream

        Returns:
            Sanic stream
        """
        q: Queue = Queue()
        task = asyncio.ensure_future(
            self.on_message_wrapper(
                on_new_message, text, q, sender_id, input_channel, metadata
            )
        )
        while True:
            result = await q.get()
            if result == "DONE":
                break
            else:
                await resp.write(json.dumps(result) + "\n")

        await task

    async def receive_messages(
        self, request: Request, on_new_message: Callable[[UserMessage], Awaitable[None]]
    ) -> Union[BaseHTTPResponse, ResponseStream]:
        sender_id = await self._extract_sender(request)
        text = self._extract_message(request)
        should_use_stream = rasa.utils.endpoints.bool_arg(
            request, "stream", default=False
        )
        input_channel = self._extract_input_channel(request)
        metadata = self.get_metadata(request)

        if should_use_stream:
            return ResponseStream(
                partial(
                    self.stream_response,
                    on_new_message,
                    text,
                    sender_id,
                    input_channel,
                    metadata,
                ),
                content_type="text/event-stream",
            )
        else:
            collector = CollectingOutputChannel()
            # noinspection PyBroadException
            try:
                await on_new_message(
                    UserMessage(
                        text,
                        collector,
                        sender_id,
                        input_channel=input_channel,
                        metadata=metadata,
                        headers=request.headers,
                    )
                )
            except CancelledError:
                structlogger.error(
                    "rest.message.received.timeout",
                    event_info="Message processing was cancelled.",
                )
            except Exception as e:
                structlogger.exception(
                    "rest.message.received.failure",
                    event_info=f"Message processing failed. Error: {e}",
                )

            return response.json(collector.messages)

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[None]]
    ) -> Blueprint:
        """Groups the collection of endpoints used by rest channel."""
        module_type = inspect.getmodule(self)
        if module_type is not None:
            module_name = module_type.__name__
        else:
            module_name = None

        custom_webhook = Blueprint(
            "custom_webhook_{}".format(type(self).__name__),
            module_name,
        )

        # noinspection PyUnusedLocal
        @custom_webhook.route("/", methods=["GET"])
        async def health(request: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @custom_webhook.route("/webhook", methods=["POST"])
        async def receive(request: Request) -> Union[ResponseStream, BaseHTTPResponse]:
            return await self.receive_messages(request, on_new_message)

        return custom_webhook


class QueueOutputChannel(CollectingOutputChannel):
    """Output channel that collects send messages in a list.

    (doesn't send them anywhere, just collects them).
    """

    # FIXME: this is breaking Liskov substitution principle
    # and would require some user-facing refactoring to address
    messages: Queue  # type: ignore[assignment]

    @classmethod
    def name(cls) -> Text:
        """Name of QueueOutputChannel."""
        return "queue"

    # noinspection PyMissingConstructor
    def __init__(self, message_queue: Optional[Queue] = None) -> None:
        super().__init__()
        self.messages = Queue() if not message_queue else message_queue

    def latest_output(self) -> NoReturn:
        raise NotImplementedError("A queue doesn't allow to peek at messages.")

    async def _persist_message(self, message: Dict[Text, Any]) -> None:
        await self.messages.put(message)



import asyncio
import json
import os
import sqlite3
from abc import ABC, abstractmethod
from contextlib import contextmanager
from datetime import datetime, timezone
from typing import Any, Callable, Dict, Generator, List, Optional, Type

import structlog

from rasa.builder.copilot.constants import (
    DEFAULT_COPILOT_CHAT_ID,
    ROLE_COPILOT,
    ROLE_COPILOT_INTERNAL,
    ROLE_SYSTEM,
    ROLE_USER,
)
from rasa.builder.copilot.exceptions import CopilotHistoryDatabaseError
from rasa.builder.copilot.models import (
    ButtonContent,
    ChatMessage,
    CodeContent,
    CommitContent,
    ContentBlock,
    ConversationKey,
    CopilotChatMessage,
    CopilotSystemMessage,
    EventContent,
    FileContent,
    InternalCopilotRequestChatMessage,
    LinkContent,
    LogContent,
    LogItem,
    LogsContent,
    ReferenceEntry,
    ReferenceItem,
    ReferencesContent,
    ResponseCategory,
    TextContent,
    UserChatMessage,
)

structlogger = structlog.get_logger()


class CopilotHistoryStore(ABC):
    """Abstract interface for storing and retrieving copilot chat history."""

    @abstractmethod
    async def get(self, key: ConversationKey) -> List[ChatMessage]:
        """Get the conversation history for the given key.

        Args:
            key: The conversation key.

        Returns:
            A list of chat messages in the conversation history.
        """
        raise NotImplementedError

    @abstractmethod
    async def append(self, key: ConversationKey, message: ChatMessage) -> None:
        """Append a message to the conversation history for the given key.

        Args:
            key: The conversation key.
            message: The chat message to append.
        """
        raise NotImplementedError

    @abstractmethod
    async def delete(self, key: ConversationKey) -> None:
        """Delete the conversation history for the given key.

        Args:
            key: The conversation key.
        """
        raise NotImplementedError


class SQLiteCopilotHistoryStore(CopilotHistoryStore):
    """SQLite-based implementation of copilot history storage."""

    def __init__(self, database_path: str) -> None:
        """Initialize the SQLite store.

        Args:
            database_path: Path to the SQLite database file.
        """
        super().__init__()
        self._database_path = database_path

        # Ensure the database directory exists
        if database_dir := os.path.dirname(database_path):
            os.makedirs(database_dir, exist_ok=True)

        self._initialize_database()

    @contextmanager
    def _database_connection(self) -> Generator[sqlite3.Connection, None, None]:
        """Context manager for SQLite database connections.

        Yields:
            SQLite connection that will be automatically closed.
        """
        connection = sqlite3.connect(self._database_path)
        try:
            yield connection
        finally:
            connection.close()

    def _initialize_database(self) -> None:
        """Create the database schema if it doesn't exist."""
        with self._database_connection() as connection:
            # Create the messages table
            connection.execute(
                """
                CREATE TABLE IF NOT EXISTS copilot_messages (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    chat_id TEXT NOT NULL,
                    role TEXT NOT NULL,
                    content_json TEXT NOT NULL,
                    response_category TEXT,
                    created_at REAL NOT NULL
                )
                """
            )

            # Create index for efficient conversation queries
            connection.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_conversation
                ON copilot_messages(chat_id, created_at)
                """
            )
            connection.commit()

    @staticmethod
    def _get_message_class_for_role(role: Optional[str]) -> Type[ChatMessage]:
        """Get the appropriate message class for a given role.

        Args:
            role: The role string (e.g., 'user', 'copilot', 'copilot_internal') or None.

        Returns:
            The corresponding message class.
        """
        if role is None:
            structlogger.warning("copilot_history_store.get_message_class.missing_role")
            return CopilotChatMessage

        if role == ROLE_SYSTEM:
            return CopilotSystemMessage
        elif role == ROLE_USER:
            return UserChatMessage
        elif role == ROLE_COPILOT:
            return CopilotChatMessage
        elif role == ROLE_COPILOT_INTERNAL:
            return InternalCopilotRequestChatMessage
        else:
            structlogger.warning(
                "copilot_history_store.get_message_class.unknown_role",
                role=role,
            )
            return CopilotChatMessage

    @staticmethod
    def _deserialize_content_block(content_dict: Dict[str, Any]) -> ContentBlock:
        """Deserialize a content block from dictionary.

        Args:
            content_dict: Dictionary containing content block data.

        Returns:
            The appropriate ContentBlock instance.
        """
        raw_type = content_dict.get("type")

        if raw_type is None:
            structlogger.warning(
                "copilot_history_store.deserialize_content.missing_type",
                content_dict=content_dict,
            )
            return TextContent(type="text", text=str(content_dict))

        if not isinstance(raw_type, str):
            structlogger.warning(
                "copilot_history_store.deserialize_content.non_string_type",
                provided_type=type(raw_type).__name__,
                value=raw_type,
            )
            return TextContent(type="text", text=str(content_dict))

        content_type = raw_type.strip().lower()

        constructors: Dict[str, Callable[..., ContentBlock]] = {
            "text": TextContent,
            "code": CodeContent,
            "file": FileContent,
            "log": LogContent,
            "link": LinkContent,
            "button": ButtonContent,
            "references": ReferencesContent,
            "logs": LogsContent,
            "commit": CommitContent,
        }

        if content_type in constructors:
            return constructors[content_type](**content_dict)

        if content_type == "event":
            # Special handling for EventContent to avoid double nesting
            if "event_data" in content_dict:
                content_copy = content_dict.copy()
                event_data = content_copy.pop("event_data", {}) or {}
                result_dict = {
                    "type": content_copy.get("type"),
                    "event": content_copy.get("event"),
                }
                result_dict.update(event_data)
                return EventContent(**result_dict)
            return EventContent(**content_dict)

        structlogger.warning(
            "copilot_history_store.deserialize_content.unknown_type",
            content_type=content_type,
        )
        return TextContent(type="text", text=str(content_dict))

    @staticmethod
    def _create_message_from_db_row(
        role: str,
        content: List,
        response_category: Optional[str],
        timestamp: Optional[float] = None,
    ) -> ChatMessage:
        """Create a chat message from database row data.

        Args:
            role: The role string (e.g., 'user', 'copilot', 'copilot_internal').
            content: The content list (will be ignored for system messages).
            response_category: Optional response category string.
            timestamp: Optional unix timestamp (UTC) when the message was created.

        Returns:
            The appropriate ChatMessage instance.
        """
        message_class = SQLiteCopilotHistoryStore._get_message_class_for_role(role)

        kwargs: Dict[str, Any] = {
            "role": role,
            "response_category": ResponseCategory(response_category)
            if response_category
            else None,
            "timestamp": timestamp,
        }

        # System messages don't have content field
        if role != ROLE_SYSTEM:
            # Deserialize content blocks properly
            deserialized_content = [
                SQLiteCopilotHistoryStore._deserialize_content_block(block)
                for block in content
            ]
            kwargs["content"] = deserialized_content

        return message_class(**kwargs)

    @staticmethod
    def _serialize_messages(messages: List[CopilotChatMessage]) -> str:
        """Serialize messages to JSON string.

        Args:
            messages: List of copilot chat messages.

        Returns:
            JSON string representation of the messages.
        """
        return json.dumps(
            [message.model_dump() for message in messages], ensure_ascii=False
        )

    @staticmethod
    def _deserialize_messages(json_payload: str) -> List[ChatMessage]:
        """Deserialize messages from JSON string.

        Args:
            json_payload: JSON string containing serialized messages.

        Returns:
            List of chat messages.
        """
        raw_messages = json.loads(json_payload or "[]")
        messages: List[ChatMessage] = []
        for message_data in raw_messages:
            role = message_data.get("role")
            message_class = SQLiteCopilotHistoryStore._get_message_class_for_role(role)
            messages.append(message_class(**message_data))
        return messages

    def _read_conversation_from_database(
        self, key: ConversationKey
    ) -> List[ChatMessage]:
        """Read conversation messages from SQLite database.

        Args:
            key: Conversation key.

        Returns:
            List of chat messages for the conversation.
        """
        with self._database_connection() as connection:
            cursor = connection.execute(
                """
                SELECT role, content_json, response_category, created_at
                FROM copilot_messages
                WHERE chat_id=?
                ORDER BY created_at
                """,
                key.to_tuple(),
            )
            messages: List[ChatMessage] = []
            for row in cursor.fetchall():
                role, content_json, response_category, created_at = row
                content = json.loads(content_json)

                message = self._create_message_from_db_row(
                    role=role,
                    content=content,
                    response_category=response_category,
                    timestamp=created_at,
                )
                messages.append(message)
            return messages

    def _write_single_message_to_database(
        self, key: ConversationKey, message: ChatMessage
    ) -> None:
        """Write a single message to SQLite database.

        Args:
            key: Conversation key.
            message: Chat message to store.
        """
        with self._database_connection() as connection:
            timestamp = (
                message.timestamp
                if message.timestamp is not None
                else datetime.now(timezone.utc).timestamp()
            )

            if message.timestamp is None:
                message.timestamp = timestamp

            # CopilotSystemMessage doesn't have a content field
            content = []
            if message.role != ROLE_SYSTEM:
                content = [c.model_dump() for c in message.content]

            content_json = json.dumps(content)

            connection.execute(
                """
                INSERT INTO copilot_messages
                (
                    chat_id,
                    role,
                    content_json,
                    response_category,
                    created_at
                )
                VALUES (?, ?, ?, ?, ?)
                """,
                (
                    *key.to_tuple(),
                    message.role,
                    content_json,
                    message.response_category.value
                    if message.response_category
                    else None,
                    timestamp,
                ),
            )
            connection.commit()

    def _delete_conversation_from_database(self, key: ConversationKey) -> None:
        """Delete conversation from SQLite database.

        Args:
            key: Conversation key.
        """
        with self._database_connection() as connection:
            connection.execute(
                """
                DELETE FROM copilot_messages
                WHERE chat_id=?
                """,
                key.to_tuple(),
            )
            connection.commit()

    def _handle_missing_table_error(self, exc: sqlite3.OperationalError) -> bool:
        """Check if error is due to missing table and reinitialize if needed.

        Args:
            exc: The SQLite operational error.

        Returns:
            True if table was missing and reinitialized, False otherwise.
        """
        if "no such table" in str(exc).lower():
            structlogger.warning("copilot_history_store.table_missing", error=str(exc))
            self._initialize_database()
            return True
        return False

    async def get(self, key: ConversationKey) -> List[ChatMessage]:
        """Get conversation history from SQLite database.

        Args:
            key: The conversation key.

        Returns:
            A list of chat messages in the conversation history.

        Raises:
            CopilotHistoryDatabaseError: If database operation fails.
        """
        try:
            return await asyncio.to_thread(self._read_conversation_from_database, key)
        except sqlite3.OperationalError as exc:
            if self._handle_missing_table_error(exc):
                return []
            raise CopilotHistoryDatabaseError(
                f"Failed to get conversation {key.chat_id}: {exc}"
            )
        except Exception as exc:
            structlogger.error(
                "copilot_history_store.get_failed",
                conversation_key=key.chat_id,
                error=str(exc),
            )
            raise CopilotHistoryDatabaseError(
                f"Failed to get conversation {key.chat_id}: {exc}"
            )

    async def append(self, key: ConversationKey, message: ChatMessage) -> None:
        """Append a message to SQLite conversation history.

        Args:
            key: The conversation key.
            message: The chat message to append.

        Raises:
            CopilotHistoryDatabaseError: If database operation fails.
        """
        try:
            await asyncio.to_thread(
                self._write_single_message_to_database,
                key,
                message,
            )
        except sqlite3.OperationalError as exc:
            if self._handle_missing_table_error(exc):
                await asyncio.to_thread(
                    self._write_single_message_to_database, key, message
                )
                return
            raise CopilotHistoryDatabaseError(
                f"Failed to append to conversation {key.chat_id}: {exc}"
            )
        except Exception as exc:
            structlogger.error(
                "copilot_history_store.append_failed",
                conversation_key=key.chat_id,
                message_role=getattr(message, "role", "unknown"),
                error=str(exc),
            )
            raise CopilotHistoryDatabaseError(
                f"Failed to append to conversation {key.chat_id}: {exc}"
            )

    async def delete(self, key: ConversationKey) -> None:
        """Delete conversation history from SQLite database.

        Args:
            key: The conversation key.

        Raises:
            CopilotHistoryDatabaseError: If database operation fails.
        """
        try:
            await asyncio.to_thread(self._delete_conversation_from_database, key)
        except Exception as exc:
            structlogger.error(
                "copilot_history_store.delete_failed",
                conversation_key=key.chat_id,
                error=str(exc),
            )
            raise CopilotHistoryDatabaseError(
                f"Failed to delete conversation {key.chat_id}: {exc}"
            )


async def persist_copilot_message_to_history(
    content: Optional[List[ContentBlock]] = None,
    text: Optional[str] = None,
    references: Optional[List[ReferenceEntry]] = None,
    chat_id: str = DEFAULT_COPILOT_CHAT_ID,
    response_category: ResponseCategory = ResponseCategory.COPILOT,
    commit: Optional[Dict[str, Any]] = None,
) -> None:
    """Persist a copilot message to conversation history.

    This is the general function for persisting any copilot message.
    You can either provide a pre-built content array, or text with optional references.

    Args:
        content: Optional list of ContentBlock objects (takes precedence over text)
        text: Optional text content for simple text-only messages
        references: Optional list of ReferenceEntry objects to include
        chat_id: The chat ID to persist the message to
        response_category: The response category for the message
        commit: Optional commit information to include
    """
    # If neither content nor text provided, nothing to persist
    if not content and not text:
        return

    try:
        # Import here to avoid circular dependency
        from rasa.builder.llm_service import llm_service

        conversation_key = ConversationKey(chat_id=chat_id)

        # Build content blocks
        if content:
            message_content = content
        else:
            message_content = [TextContent(type="text", text=text)]

            # Add references as a content block if provided
            if references:
                reference_items = [
                    ReferenceItem(index=ref.index, title=ref.title, url=ref.url)
                    for ref in references
                ]
                message_content.append(
                    ReferencesContent(type="references", references=reference_items)
                )

            # Add commit as a content block if provided
            if commit:
                message_content.append(CommitContent(type="commit", commit=commit))

        copilot_message = CopilotChatMessage(
            role="copilot",
            content=message_content,
            response_category=response_category,
        )

        await llm_service.history_store.append(conversation_key, copilot_message)

        structlogger.debug(
            "copilot_message_persisted",
            chat_id=chat_id,
            response_category=response_category.value,
            content_blocks=len(message_content),
        )
    except Exception as persist_exc:
        structlogger.error(
            "copilot_message_persist_failed",
            error=str(persist_exc),
            chat_id=chat_id,
        )


async def persist_training_error_analysis_to_history(
    text: Optional[str] = None,
    logs: Optional[List[LogContent]] = None,
    references: Optional[List[ReferenceEntry]] = None,
    chat_id: str = DEFAULT_COPILOT_CHAT_ID,
    response_category: ResponseCategory = ResponseCategory.TRAINING_ERROR_LOG_ANALYSIS,
    commit: Optional[Dict[str, Any]] = None,
) -> None:
    """Persist training error analysis as a single message with all content types.

    Args:
        text: Optional main analysis text
        logs: Optional list of log content blocks
        references: Optional list of reference entries
        chat_id: The chat ID to persist the message to
        response_category: The response category for the message
        commit: Optional commit information to include
    """
    if not text and not logs and not references:
        return

    # Build content array with all content types
    content: List[ContentBlock] = []

    if text:
        content.append(TextContent(type="text", text=text))

    if logs:
        log_items = [
            LogItem(
                type="log",
                content=log.content,
                context=log.context,
                metadata=log.metadata or {},
            )
            for log in logs
        ]
        content.append(LogsContent(type="logs", logs=log_items))

    if references:
        reference_items = [
            ReferenceItem(index=ref.index, title=ref.title, url=ref.url)
            for ref in references
        ]
        content.append(ReferencesContent(type="references", references=reference_items))

    if commit:
        content.append(CommitContent(type="commit", commit=commit))

    await persist_copilot_message_to_history(
        content=content,
        chat_id=chat_id,
        response_category=response_category,
    )


async def persist_user_message_to_history(
    text: str,
    chat_id: str = DEFAULT_COPILOT_CHAT_ID,
) -> None:
    """Persist a user message to conversation history.

    Args:
        text: The text content of the user message
        chat_id: The chat ID to persist the message to
    """
    if not text:
        return

    try:
        # Import here to avoid circular dependency
        from rasa.builder.llm_service import llm_service

        conversation_key = ConversationKey(chat_id=chat_id)

        user_message = UserChatMessage(
            role=ROLE_USER,
            content=[TextContent(type="text", text=text)],
        )

        await llm_service.history_store.append(conversation_key, user_message)

        structlogger.debug(
            "user_message_persisted",
            chat_id=chat_id,
        )
    except Exception as persist_exc:
        structlogger.error(
            "user_message_persist_failed",
            error=str(persist_exc),
            chat_id=chat_id,
        )



from typing import Dict, List, Literal, Optional

import structlog

from rasa.builder.copilot.models import ResponseCategory
from rasa.builder.evaluator.response_classification.constants import (
    MACRO_AVERAGING_METHOD,
    MICRO_AVERAGING_METHOD,
    WEIGHTED_AVERAGING_METHOD,
)
from rasa.builder.evaluator.response_classification.models import (
    ClassificationResult,
    MetricsSummary,
    OverallClassificationMetrics,
    PerClassMetrics,
)

structlogger = structlog.get_logger()


class ResponseClassificationEvaluator:
    def __init__(self):  # type: ignore[no-untyped-def]
        self._classes: List[ResponseCategory] = [
            ResponseCategory.COPILOT,
            ResponseCategory.OUT_OF_SCOPE_DETECTION,
            ResponseCategory.ROLEPLAY_DETECTION,
            ResponseCategory.KNOWLEDGE_BASE_ACCESS_REQUESTED,
            ResponseCategory.ERROR_FALLBACK,
            # TODO: Add the greetings and goodbyes as support once the orchestrator
            #       aproach is implemented
        ]
        self._true_positives_per_class: Dict[ResponseCategory, int] = {
            clazz: 0 for clazz in self._classes
        }
        self._false_positives_per_class: Dict[ResponseCategory, int] = {
            clazz: 0 for clazz in self._classes
        }
        self._false_negatives_per_class: Dict[ResponseCategory, int] = {
            clazz: 0 for clazz in self._classes
        }
        self._support_per_class: Dict[ResponseCategory, int] = {
            clazz: 0 for clazz in self._classes
        }

        self._evaluated = False

    @property
    def metrics_summary(self) -> Optional[MetricsSummary]:
        """Get the metrics summary.

        Returns:
            MetricsSummary with structured per-class and overall metrics if
            the evaluator has been run on the data, otherwise None.
        """
        if not self._evaluated:
            structlogger.warning(
                "evaluator.response_classification_evaluator"
                ".metrics_summary.not_evaluated",
                event_info="Evaluator not evaluated. Returning empty metrics summary.",
            )
            return None

        return self._get_metrics_summary()

    def reset(self) -> None:
        self._true_positives_per_class = {clazz: 0 for clazz in self._classes}
        self._false_positives_per_class = {clazz: 0 for clazz in self._classes}
        self._false_negatives_per_class = {clazz: 0 for clazz in self._classes}
        self._support_per_class = {clazz: 0 for clazz in self._classes}
        self._evaluated = False

    def evaluate(self, item_results: List[ClassificationResult]) -> MetricsSummary:
        """Evaluate the classifier on the given item results."""
        if self._evaluated:
            structlogger.warning(
                "evaluator.response_classification_evaluator.evaluate.already_evaluated",
                event_info="Evaluator already evaluated. Resetting evaluator.",
            )
            self.reset()

        for result in item_results:
            # Skip and raise a warning if the class is not in the list of classes
            if result.expected not in self._classes:
                structlogger.warning(
                    "evaluator.response_classification_evaluator"
                    ".evaluate.class_not_recognized",
                    event_info=(
                        f"Class '{result.expected}' is not recognized. "
                        f"Skipping evaluation for this class."
                    ),
                    expected_class=result.expected,
                    classes=self._classes,
                )
                continue

            # Update support for the expected class
            if result.expected in self._support_per_class:
                self._support_per_class[result.expected] += 1

            # Calculate TP, FP, FN per class
            for clazz in self._classes:
                if result.prediction == clazz and result.expected == clazz:
                    self._true_positives_per_class[clazz] += 1

                elif result.prediction == clazz and result.expected != clazz:
                    self._false_positives_per_class[clazz] += 1

                elif result.prediction != clazz and result.expected == clazz:
                    self._false_negatives_per_class[clazz] += 1

        self._evaluated = True
        return self._get_metrics_summary()

    def calculate_precision_per_class(self, clazz: ResponseCategory) -> float:
        """Calculate precision for a specific response category."""
        tp = self._true_positives_per_class.get(clazz, 0)
        fp = self._false_positives_per_class.get(clazz, 0)

        if tp + fp == 0:
            return 0.0

        return tp / (tp + fp)

    def calculate_recall_per_class(self, clazz: ResponseCategory) -> float:
        """Calculate recall for a specific response category."""
        tp = self._true_positives_per_class.get(clazz, 0)
        fn = self._false_negatives_per_class.get(clazz, 0)

        if tp + fn == 0:
            return 0.0

        return tp / (tp + fn)

    def calculate_f1_per_class(self, clazz: ResponseCategory) -> float:
        """Calculate F1 score for a specific response category."""
        precision = self.calculate_precision_per_class(clazz)
        recall = self.calculate_recall_per_class(clazz)

        if precision + recall == 0:
            return 0.0

        return 2 * (precision * recall) / (precision + recall)

    def calculate_precision(
        self, average: Literal["micro", "macro", "weighted"] = MICRO_AVERAGING_METHOD
    ) -> float:
        """Calculate precision with specified averaging method."""
        if average == MICRO_AVERAGING_METHOD:
            return self._calculate_micro_precision()
        elif average == MACRO_AVERAGING_METHOD:
            return self._calculate_macro_precision()
        elif average == WEIGHTED_AVERAGING_METHOD:
            return self._calculate_weighted_avg_precision()
        else:
            raise ValueError(f"Invalid averaging method: {average}")

    def _calculate_micro_precision(self) -> float:
        """Calculate overall precision with specified averaging method.

        Calculates the metric globally by aggregating the total true positives, false
        positives, across all classes. Each sample contributes equally to the final
        score.
        """
        total_tp = sum(self._true_positives_per_class.values())
        total_fp = sum(self._false_positives_per_class.values())

        if total_tp + total_fp == 0:
            return 0.0

        return total_tp / (total_tp + total_fp)

    def _calculate_macro_precision(self) -> float:
        """Calculate macro-averaged precision.

        Calculates the metric independently for each class and then takes the
        unweighted average. Each class contributes equally.
        """
        precisions = [
            self.calculate_precision_per_class(clazz) for clazz in self._classes
        ]
        return sum(precisions) / len(precisions) if precisions else 0.0

    def _calculate_weighted_avg_precision(self) -> float:
        """Calculate weighted-averaged precision.

        Calculates the metric independently for each class and then takes the average
        weighted by the class support (number of true samples per class).
        """
        total_support = sum(self._support_per_class.values())
        if total_support == 0:
            return 0.0

        weighted_sum = 0.0
        for clazz in self._classes:
            precision = self.calculate_precision_per_class(clazz)
            support = self._support_per_class.get(clazz, 0)
            weighted_sum += precision * support

        return weighted_sum / total_support

    def calculate_recall(
        self, average: Literal["micro", "macro", "weighted"] = MICRO_AVERAGING_METHOD
    ) -> float:
        """Calculate recall with specified averaging method."""
        if average == MICRO_AVERAGING_METHOD:
            return self._calculate_micro_recall()
        elif average == MACRO_AVERAGING_METHOD:
            return self._calculate_macro_recall()
        elif average == WEIGHTED_AVERAGING_METHOD:
            return self._calculate_weighted_avg_recall()
        else:
            raise ValueError(f"Invalid averaging method: {average}")

    def _calculate_micro_recall(self) -> float:
        """Calculate micro-averaged recall.

        Calculates the metric globally by aggregating the total true positives, false
        negatives, across all classes. Each sample contributes equally to the final
        score.
        """
        total_tp = sum(self._true_positives_per_class.values())
        total_fn = sum(self._false_negatives_per_class.values())

        if total_tp + total_fn == 0:
            return 0.0

        return total_tp / (total_tp + total_fn)

    def _calculate_macro_recall(self) -> float:
        """Calculate macro-averaged recall.

        Calculates the metric independently for each class and then takes the
        unweighted average. Each class contributes equally.
        """
        recalls = [self.calculate_recall_per_class(clazz) for clazz in self._classes]
        return sum(recalls) / len(recalls) if recalls else 0.0

    def _calculate_weighted_avg_recall(self) -> float:
        """Calculate weighted-averaged recall.

        Calculates the metric independently for each class and then takes the average
        weighted by the class support (number of true samples per class).
        """
        total_support = sum(self._support_per_class.values())
        if total_support == 0:
            return 0.0

        weighted_sum = 0.0
        for clazz in self._classes:
            recall = self.calculate_recall_per_class(clazz)
            support = self._support_per_class.get(clazz, 0)
            weighted_sum += recall * support

        return weighted_sum / total_support

    def calculate_f1(
        self, average: Literal["micro", "macro", "weighted"] = MICRO_AVERAGING_METHOD
    ) -> float:
        """Calculate F1 score with specified averaging method."""
        if average == MICRO_AVERAGING_METHOD:
            return self._calculate_micro_f1()
        elif average == MACRO_AVERAGING_METHOD:
            return self._calculate_macro_f1()
        elif average == WEIGHTED_AVERAGING_METHOD:
            return self._calculate_weighted_avg_f1()
        else:
            raise ValueError(f"Invalid averaging method: {average}")

    def _calculate_micro_f1(self) -> float:
        """Calculate micro-averaged F1 score.

        Calculates the metric globally by aggregating the total true positives, false
        positives, and false negatives across all classes. Each sample contributes
        equally to the final score.
        """
        micro_precision = self._calculate_micro_precision()
        micro_recall = self._calculate_micro_recall()

        if micro_precision + micro_recall == 0:
            return 0.0

        return 2 * (micro_precision * micro_recall) / (micro_precision + micro_recall)

    def _calculate_macro_f1(self) -> float:
        """Calculate macro-averaged F1 score.

        Calculates the metric independently for each class and then takes the
        unweighted average. Each class contributes equally.
        """
        f1_scores = [self.calculate_f1_per_class(clazz) for clazz in self._classes]
        return sum(f1_scores) / len(f1_scores) if f1_scores else 0.0

    def _calculate_weighted_avg_f1(self) -> float:
        """Calculate weighted F1 score.

        Calculates the metric independently for each class and then takes the average
        weighted by the class support (number of true samples per class).
        """
        total_support = sum(self._support_per_class.values())
        if total_support == 0:
            return 0.0

        weighted_sum = 0.0
        for clazz in self._classes:
            f1 = self.calculate_f1_per_class(clazz)
            support = self._support_per_class.get(clazz, 0)
            weighted_sum += f1 * support

        return weighted_sum / total_support

    def _get_metrics_summary(self) -> MetricsSummary:
        """Get the metrics summary without Optional wrapper.

        This method assumes the evaluator has been evaluated and will always
        return a MetricsSummary.
        """
        # Build per-class metrics
        per_class_metrics: Dict[ResponseCategory, PerClassMetrics] = {}
        for clazz in self._classes:
            per_class_metrics[clazz] = PerClassMetrics(
                precision=self.calculate_precision_per_class(clazz),
                recall=self.calculate_recall_per_class(clazz),
                f1=self.calculate_f1_per_class(clazz),
                support=self._support_per_class.get(clazz, 0),
                true_positives=self._true_positives_per_class.get(clazz, 0),
                false_positives=self._false_positives_per_class.get(clazz, 0),
                false_negatives=self._false_negatives_per_class.get(clazz, 0),
            )

        # Build overall metrics
        overall_metrics = OverallClassificationMetrics(
            micro_precision=self.calculate_precision(MICRO_AVERAGING_METHOD),
            macro_precision=self.calculate_precision(MACRO_AVERAGING_METHOD),
            weighted_avg_precision=self.calculate_precision(WEIGHTED_AVERAGING_METHOD),
            micro_recall=self.calculate_recall(MICRO_AVERAGING_METHOD),
            macro_recall=self.calculate_recall(MACRO_AVERAGING_METHOD),
            weighted_avg_recall=self.calculate_recall(WEIGHTED_AVERAGING_METHOD),
            micro_f1=self.calculate_f1(MICRO_AVERAGING_METHOD),
            macro_f1=self.calculate_f1(MACRO_AVERAGING_METHOD),
            weighted_avg_f1=self.calculate_f1(WEIGHTED_AVERAGING_METHOD),
            support=sum(self._support_per_class.values()),
            true_positives=sum(self._true_positives_per_class.values()),
            false_positives=sum(self._false_positives_per_class.values()),
            false_negatives=sum(self._false_negatives_per_class.values()),
        )
        return MetricsSummary(per_class=per_class_metrics, overall=overall_metrics)



from __future__ import annotations

import base64
import copy
import json
import logging
import zlib
from pathlib import Path
from typing import Any, Dict, List, Optional, Text

import structlog
from tqdm import tqdm

import rasa.shared.utils.io
import rasa.utils.io
from rasa.core.constants import (
    DEFAULT_MAX_HISTORY,
    MEMOIZATION_POLICY_PRIORITY,
    POLICY_MAX_HISTORY,
    POLICY_PRIORITY,
)
from rasa.core.featurizers.tracker_featurizers import (
    FEATURIZER_FILE,
    MaxHistoryTrackerFeaturizer,
    TrackerFeaturizer,
)
from rasa.core.policies.policy import Policy, PolicyPrediction, SupportedData
from rasa.engine.graph import ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.constants import ACTION_LISTEN_NAME
from rasa.shared.core.domain import Domain, State
from rasa.shared.core.events import ActionExecuted
from rasa.shared.core.generator import TrackerWithCachedStates
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import FileIOException
from rasa.shared.utils.io import is_logging_disabled

logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.POLICY_WITHOUT_END_TO_END_SUPPORT, is_trainable=True
)
class MemoizationPolicy(Policy):
    """A policy that follows exact examples of `max_history` turns in training stories.

    Since `slots` that are set some time in the past are
    preserved in all future feature vectors until they are set
    to None, this policy implicitly remembers and most importantly
    recalls examples in the context of the current dialogue
    longer than `max_history`.

    This policy is not supposed to be the only policy in an ensemble,
    it is optimized for precision and not recall.
    It should get a 100% precision because it emits probabilities of 1.1
    along it's predictions, which makes every mistake fatal as
    no other policy can overrule it.

    If it is needed to recall turns from training dialogues where
    some slots might not be set during prediction time, and there are
    training stories for this, use AugmentedMemoizationPolicy.
    """

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """Returns the default config (see parent class for full docstring)."""
        # please make sure to update the docs when changing a default parameter
        return {
            "enable_feature_string_compression": True,
            "use_nlu_confidence_as_score": False,
            POLICY_PRIORITY: MEMOIZATION_POLICY_PRIORITY,
            POLICY_MAX_HISTORY: DEFAULT_MAX_HISTORY,
        }

    def _standard_featurizer(self) -> MaxHistoryTrackerFeaturizer:
        # Memoization policy always uses MaxHistoryTrackerFeaturizer
        # without state_featurizer
        return MaxHistoryTrackerFeaturizer(
            state_featurizer=None, max_history=self.config[POLICY_MAX_HISTORY]
        )

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        featurizer: Optional[TrackerFeaturizer] = None,
        lookup: Optional[Dict] = None,
    ) -> None:
        """Initialize the policy."""
        super().__init__(config, model_storage, resource, execution_context, featurizer)
        self.lookup = lookup or {}

    def _create_lookup_from_states(
        self,
        trackers_as_states: List[List[State]],
        trackers_as_actions: List[List[Text]],
    ) -> Dict[Text, Text]:
        """Creates lookup dictionary from the tracker represented as states.

        Args:
            trackers_as_states: representation of the trackers as a list of states
            trackers_as_actions: representation of the trackers as a list of actions

        Returns:
            lookup dictionary
        """
        lookup: Dict[Text, Text] = {}

        if not trackers_as_states:
            return lookup

        assert len(trackers_as_actions[0]) == 1, (
            f"The second dimension of trackers_as_action should be 1, "
            f"instead of {len(trackers_as_actions[0])}"
        )

        ambiguous_feature_keys = set()

        pbar = tqdm(
            zip(trackers_as_states, trackers_as_actions),
            desc="Processed actions",
            disable=is_logging_disabled(),
        )
        for states, actions in pbar:
            action = actions[0]

            feature_key = self._create_feature_key(states)
            if not feature_key:
                continue

            if feature_key not in ambiguous_feature_keys:
                if feature_key in lookup.keys():
                    if lookup[feature_key] != action:
                        # delete contradicting example created by
                        # partial history augmentation from memory
                        ambiguous_feature_keys.add(feature_key)
                        del lookup[feature_key]
                else:
                    lookup[feature_key] = action
            pbar.set_postfix({"# examples": "{:d}".format(len(lookup))})

        return lookup

    def _create_feature_key(self, states: List[State]) -> Optional[Text]:
        if not states:
            return None

        # we sort keys to make sure that the same states
        # represented as dictionaries have the same json strings
        # quotes are removed for aesthetic reasons
        feature_str = json.dumps(states, sort_keys=True).replace('"', "")
        if self.config["enable_feature_string_compression"]:
            compressed = zlib.compress(
                bytes(feature_str, rasa.shared.utils.io.DEFAULT_ENCODING)
            )
            return base64.b64encode(compressed).decode(
                rasa.shared.utils.io.DEFAULT_ENCODING
            )
        else:
            return feature_str

    def train(
        self,
        training_trackers: List[TrackerWithCachedStates],
        domain: Domain,
        **kwargs: Any,
    ) -> Resource:
        # only considers original trackers (no augmented ones)
        training_trackers = [
            t
            for t in training_trackers
            if not hasattr(t, "is_augmented") or not t.is_augmented
        ]
        training_trackers = SupportedData.trackers_for_supported_data(
            self.supported_data(), training_trackers
        )

        (
            trackers_as_states,
            trackers_as_actions,
        ) = self.featurizer.training_states_and_labels(training_trackers, domain)
        self.lookup = self._create_lookup_from_states(
            trackers_as_states, trackers_as_actions
        )
        logger.debug(f"Memorized {len(self.lookup)} unique examples.")

        self.persist()
        return self._resource

    def _recall_states(self, states: List[State]) -> Optional[Text]:
        return self.lookup.get(self._create_feature_key(states))

    def recall(
        self,
        states: List[State],
        tracker: DialogueStateTracker,
        domain: Domain,
        rule_only_data: Optional[Dict[Text, Any]],
    ) -> Optional[Text]:
        """Finds the action based on the given states.

        Args:
            states: List of states.
            tracker: The tracker.
            domain: The Domain.
            rule_only_data: Slots and loops which are specific to rules and hence
                should be ignored by this policy.

        Returns:
            The name of the action.
        """
        return self._recall_states(states)

    def _prediction_result(
        self, action_name: Text, tracker: DialogueStateTracker, domain: Domain
    ) -> List[float]:
        result = self._default_predictions(domain)
        if action_name:
            if (
                self.config["use_nlu_confidence_as_score"]
                and tracker.latest_message is not None
            ):
                # the memoization will use the confidence of NLU on the latest
                # user message to set the confidence of the action
                score = tracker.latest_message.intent.get("confidence", 1.0)
            else:
                score = 1.0

            result[domain.index_for_action(action_name)] = score

        return result

    async def predict_action_probabilities(
        self,
        tracker: DialogueStateTracker,
        domain: Domain,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        **kwargs: Any,
    ) -> PolicyPrediction:
        """Predicts the next action the bot should take after seeing the tracker.

        Args:
            tracker: the :class:`rasa.core.trackers.DialogueStateTracker`
            domain: the :class:`rasa.shared.core.domain.Domain`
            rule_only_data: Slots and loops which are specific to rules and hence
                should be ignored by this policy.
            **kwargs: Additional arguments.

        Returns:
             The policy's prediction (e.g. the probabilities for the actions).
        """
        result = self._default_predictions(domain)
        if self.should_abstain_in_coexistence(tracker, False):
            return self._prediction(result)

        states = self._prediction_states(tracker, domain, rule_only_data=rule_only_data)
        structlogger.debug(
            "memoization.predict.actions", tracker_states=copy.deepcopy(states)
        )
        predicted_action_name = self.recall(
            states, tracker, domain, rule_only_data=rule_only_data
        )
        if predicted_action_name is not None:
            logger.debug(f"There is a memorised next action '{predicted_action_name}'")
            result = self._prediction_result(predicted_action_name, tracker, domain)
        else:
            logger.debug("There is no memorised next action")

        return self._prediction(result)

    def _metadata(self) -> Dict[Text, Any]:
        return {"lookup": self.lookup}

    @classmethod
    def _metadata_filename(cls) -> Text:
        return "memorized_turns.json"

    def persist(self) -> None:
        """Persists the policy to storage."""
        with self._model_storage.write_to(self._resource) as path:
            # not all policies have a featurizer
            if self.featurizer is not None:
                self.featurizer.persist(path)

            file = Path(path) / self._metadata_filename()

            rasa.shared.utils.io.create_directory_for_file(file)
            rasa.shared.utils.io.dump_obj_as_json_to_file(file, self._metadata())

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> MemoizationPolicy:
        """Loads a trained policy (see parent class for full docstring)."""
        featurizer = None
        lookup = None

        try:
            with model_storage.read_from(resource) as path:
                metadata_file = Path(path) / cls._metadata_filename()
                metadata = rasa.shared.utils.io.read_json_file(metadata_file)
                lookup = metadata["lookup"]

                if (Path(path) / FEATURIZER_FILE).is_file():
                    featurizer = TrackerFeaturizer.load(path)

        except (ValueError, FileNotFoundError, FileIOException):
            logger.warning(
                f"Couldn't load metadata for policy '{cls.__name__}' as the persisted "
                f"metadata couldn't be loaded."
            )

        return cls(
            config,
            model_storage,
            resource,
            execution_context,
            featurizer=featurizer,
            lookup=lookup,
        )


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.POLICY_WITHOUT_END_TO_END_SUPPORT, is_trainable=True
)
class AugmentedMemoizationPolicy(MemoizationPolicy):
    """The policy that remembers examples from training stories for `max_history` turns.

    If it is needed to recall turns from training dialogues
    where some slots might not be set during prediction time,
    add relevant stories without such slots to training data.
    E.g. reminder stories.

    Since `slots` that are set some time in the past are
    preserved in all future feature vectors until they are set
    to None, this policy has a capability to recall the turns
    up to `max_history` from training stories during prediction
    even if additional slots were filled in the past
    for current dialogue.
    """

    @staticmethod
    def _strip_leading_events_until_action_executed(
        tracker: DialogueStateTracker, again: bool = False
    ) -> Optional[DialogueStateTracker]:
        """Truncates the tracker to begin at the next `ActionExecuted` event.

        Args:
            tracker: The tracker to truncate.
            again: When true, truncate tracker at the second action.
                Otherwise truncate to the first action.

        Returns:
            The truncated tracker if there were actions present.
            If none are found, returns `None`.
        """
        idx_of_first_action = None
        idx_of_second_action = None

        applied_events = tracker.applied_events(True)

        # we need to find second executed action
        for e_i, event in enumerate(applied_events):
            if isinstance(event, ActionExecuted):
                if idx_of_first_action is None:
                    idx_of_first_action = e_i
                else:
                    idx_of_second_action = e_i
                    break

        # use first action, if we went first time and second action, if we went again
        idx_to_use = idx_of_second_action if again else idx_of_first_action
        if idx_to_use is None:
            return None

        # make second ActionExecuted the first one
        events = applied_events[idx_to_use:]
        if not events:
            return None

        truncated_tracker = tracker.init_copy()
        for e in events:
            truncated_tracker.update(e)

        return truncated_tracker

    def _recall_using_truncation(
        self,
        old_states: List[State],
        tracker: DialogueStateTracker,
        domain: Domain,
        rule_only_data: Optional[Dict[Text, Any]],
    ) -> Optional[Text]:
        """Attempts to match memorized states to progressively shorter trackers.

        This method iteratively removes the oldest events up to the next action
        executed and checks if the truncated event sequence matches some memorized
        states, until a match has been found or until the even sequence has been
        exhausted.

        Args:
            old_states: List of states.
            tracker: The tracker.
            domain: The Domain.
            rule_only_data: Slots and loops which are specific to rules and hence
                should be ignored by this policy.

        Returns:
            The name of the action.
        """
        logger.debug("Launch DeLorean...")

        # Truncate the tracker based on `max_history`
        truncated_tracker: Optional[DialogueStateTracker] = (
            _trim_tracker_by_max_history(tracker, self.config[POLICY_MAX_HISTORY])
        )
        truncated_tracker = self._strip_leading_events_until_action_executed(
            truncated_tracker
        )
        while truncated_tracker is not None:
            states = self._prediction_states(
                truncated_tracker, domain, rule_only_data=rule_only_data
            )

            if old_states != states:
                # check if we like new futures
                memorised = self._recall_states(states)
                if memorised is not None:
                    structlogger.debug(
                        "memoization.states_recall", states=copy.deepcopy(states)
                    )
                    return memorised
                old_states = states

            # go back again
            truncated_tracker = self._strip_leading_events_until_action_executed(
                truncated_tracker, again=True
            )

        # No match found
        structlogger.debug(
            "memoization.states_recall", old_states=copy.deepcopy(old_states)
        )
        return None

    def recall(
        self,
        states: List[State],
        tracker: DialogueStateTracker,
        domain: Domain,
        rule_only_data: Optional[Dict[Text, Any]],
    ) -> Optional[Text]:
        """Finds the action based on the given states.

        Uses back to the future idea to change the past and check whether the new future
        can be used to recall the action.

        Args:
            states: List of states.
            tracker: The tracker.
            domain: The Domain.
            rule_only_data: Slots and loops which are specific to rules and hence
                should be ignored by this policy.

        Returns:
            The name of the action.
        """
        predicted_action_name = self._recall_states(states)
        if predicted_action_name is None:
            # let's try a different method to recall that tracker
            return self._recall_using_truncation(
                states, tracker, domain, rule_only_data=rule_only_data
            )
        else:
            return predicted_action_name


def _get_max_applied_events_for_max_history(
    tracker: DialogueStateTracker, max_history: Optional[int]
) -> Optional[int]:
    """Computes the number of events in the tracker that correspond to max_history.

    To ensure that the last user utterance is correctly included in the prediction
    states, return the index of the most recent `action_listen` event occuring
    before the tracker would be truncated according to the value of `max_history`.

    Args:
        tracker: Some tracker holding the events
        max_history: The number of actions to count

    Returns:
        The number of events, as counted from the end of the event list, that should
        be taken into accout according to the `max_history` setting. If all events
        should be taken into account, the return value is `None`.
    """
    if not max_history:
        return None
    num_events = 0
    num_actions = 0
    for event in reversed(tracker.applied_events(True)):
        num_events += 1
        if isinstance(event, ActionExecuted):
            num_actions += 1
            if num_actions > max_history and event.action_name == ACTION_LISTEN_NAME:
                return num_events
    return None


def _trim_tracker_by_max_history(
    tracker: DialogueStateTracker, max_history: Optional[int]
) -> DialogueStateTracker:
    """Removes events from the tracker until it has `max_history` actions.

    Args:
        tracker: Some tracker.
        max_history: Number of actions to keep.

    Returns:
        A new tracker with up to `max_history` actions, or the same tracker if
        `max_history` is `None`.
    """
    max_applied_events = _get_max_applied_events_for_max_history(tracker, max_history)
    if not max_applied_events:
        return tracker

    applied_events = tracker.applied_events(True)[-max_applied_events:]
    new_tracker = tracker.init_copy()
    for event in applied_events:
        new_tracker.update(event)
    return new_tracker



from dataclasses import dataclass
from typing import Any, Dict, Text

from langcodes import Language as LangcodesLanguage
from langcodes import standardize_tag
from langcodes.tag_parser import LanguageTagError

from rasa.shared.exceptions import RasaException

CUSTOM_LANGUAGE_CODE_PREFIX = "x-"


@dataclass(frozen=True)
class Language:
    code: str
    label: str
    is_default: bool

    @classmethod
    def from_language_code(
        cls, language_code: str, is_default: bool = False
    ) -> "Language":
        """Creates a Language object from a language code.

        Args:
            language_code: The language code.
            is_default: Whether the language is the default language.

        Returns:
            A Language object.

        Raises:
            RasaException: If the language code or custom language code is invalid.
        """
        if cls.is_custom_language_code(language_code):
            cls.validate_custom_language_code(language_code)
        elif not cls.is_language_code_bcp_47_standard(language_code):
            raise RasaException(
                f"Language '{language_code}' is not a BCP 47 standard language code."
            )

        language = LangcodesLanguage.get(language_code)
        return cls(
            code=language_code,
            label=cls.get_language_label(language),
            is_default=is_default,
        )

    @staticmethod
    def is_language_code_bcp_47_standard(language_code: str) -> bool:
        """Checks if a language code is a BCP 47 standard language code.

        Args:
            language_code: The language code to check.

        Returns:
            `True` if the language code is a BCP 47 standard, `False` otherwise.
        """
        try:
            standardized_language_code = standardize_tag(language_code)
            return (
                standardized_language_code == language_code
                and LangcodesLanguage.get(language_code).is_valid()
            )
        except LanguageTagError:
            return False

    @staticmethod
    def is_custom_language_code(language_code: str) -> bool:
        """Checks if a language code is a custom language code.

        Args:
            language_code: The language code to check.

        Returns:
            `True` if the language code is a custom language code, `False` otherwise.
        """
        return language_code.startswith(CUSTOM_LANGUAGE_CODE_PREFIX)

    @classmethod
    def get_language_label(cls, language: LangcodesLanguage) -> str:
        """Gets the display name of a language.

        For custom languages (in the format "x-<base_lang>-<custom_label>"),
        the label is derived from the base language code.
        This method considers that the language code has previously been validated.

        Args:
            language: The language code.

        Returns:
            The display name of the language.
        """
        language_code = str(language)

        if cls.is_custom_language_code(language_code):
            # If it's a custom language, derive the label from the base language code.
            without_prefix = language_code[len(CUSTOM_LANGUAGE_CODE_PREFIX) :]
            base_language_code, _ = without_prefix.rsplit("-", 1)
            base_language = LangcodesLanguage.get(base_language_code)
            return base_language.display_name()
        else:
            return language.display_name()

    @classmethod
    def validate_language(cls, language: LangcodesLanguage) -> None:
        """Validates a language code.

        Args:
            language: The language object to validate.

        Raises:
            RasaException: If the language validation fails.
        """
        if not language.is_valid():
            raise RasaException(f"Language '{language}' is not a valid language code.")

        language_code = str(language)
        if language_code.startswith(CUSTOM_LANGUAGE_CODE_PREFIX):
            cls.validate_custom_language_code(language_code)

    @classmethod
    def validate_custom_language_code(cls, custom_language_code: str) -> None:
        """Validates a custom language code.

        A valid custom language code should adhere to the format:
          "x-<existing_language_code>-<custom_label>"
        Example: x-en-formal or x-en-US-formal.

        Args:
            custom_language_code: The custom language code to validate.

        Raises:
            RasaException: If the custom language code validation fails.
        """
        # Ensure the custom language code starts with the custom prefix.
        if not custom_language_code.startswith(CUSTOM_LANGUAGE_CODE_PREFIX):
            raise RasaException(
                f"Custom language '{custom_language_code}' must "
                f"start with '{CUSTOM_LANGUAGE_CODE_PREFIX}'."
            )

        # Remove the custom prefix.
        without_prefix = custom_language_code[len(CUSTOM_LANGUAGE_CODE_PREFIX) :]
        if "-" not in without_prefix:
            raise RasaException(
                f"Custom language '{custom_language_code}' must be in the format "
                f"'{CUSTOM_LANGUAGE_CODE_PREFIX}<language_code>-<custom_label>'."
            )

        base_language_code, custom_label = without_prefix.rsplit("-", 1)
        if not base_language_code:
            raise RasaException(
                f"Base language in '{custom_language_code}' cannot be empty. "
                f"Expected custom language code format is "
                f"'{CUSTOM_LANGUAGE_CODE_PREFIX}<language_code>-<custom_label>'."
            )
        if not custom_label:
            raise RasaException(
                f"Custom label in '{custom_language_code}' cannot be empty."
                f"Expected custom language code format is "
                f"'{CUSTOM_LANGUAGE_CODE_PREFIX}<language_code>-<custom_label>'."
            )

        # Validate the base language code using langcodes.
        if not cls.is_language_code_bcp_47_standard(base_language_code):
            raise RasaException(
                f"Base language '{base_language_code}' in custom language "
                f"'{custom_language_code}' is not a valid language code."
            )

    def as_dict(self) -> Dict[Text, Any]:
        """Converts the Language dataclass instance into a dictionary.

        Returns:
            A dictionary representing the Language object.
        """
        return {
            "code": self.code,
            "label": self.label,
            "is_default": self.is_default,
        }



import subprocess
import tarfile
from pathlib import Path
from typing import Any, Dict, List, Optional

import structlog
from sanic import Sanic

from rasa.builder import config
from rasa.builder.constants import (
    MAX_ARCHIVE_FILE_SIZE,
    MAX_ARCHIVE_FILES,
    MAX_ARCHIVE_TOTAL_SIZE,
)
from rasa.builder.copilot.constants import (
    PROMPT_TO_BOT_KEY,
)
from rasa.builder.copilot.copilot_templated_message_provider import (
    load_copilot_handler_default_responses,
    load_copilot_template_prompts,
    load_copilot_welcome_messages,
)
from rasa.builder.copilot.history_store import (
    persist_copilot_message_to_history,
    persist_training_error_analysis_to_history,
    persist_user_message_to_history,
)
from rasa.builder.copilot.models import (
    CopilotContext,
    FileContent,
    InternalCopilotRequestChatMessage,
    LogContent,
    ResponseCategory,
    TrainingErrorLog,
)
from rasa.builder.download import download_backup_from_url
from rasa.builder.exceptions import (
    LLMGenerationError,
    ProjectGenerationError,
    TrainingError,
    ValidationError,
)
from rasa.builder.git_service import GitOperationInProgressError
from rasa.builder.job_helpers import (
    handle_revert_error,
    handle_rollback_error,
    load_or_train_agent_for_commit,
    perform_revert,
    perform_rollback,
    push_error_and_start_copilot_analysis,
    push_job_status_event,
    train_and_load_and_link_agent,
)
from rasa.builder.job_manager import JobInfo, job_manager
from rasa.builder.llm_service import llm_service
from rasa.builder.models import (
    GitCommitInfo,
    JobStatus,
)
from rasa.builder.project_generator import ProjectGenerator
from rasa.builder.training_service import (
    try_load_existing_agent,
    update_agent,
)
from rasa.builder.validation_service import validate_project
from rasa.cli.scaffold import ProjectTemplateName

structlogger = structlog.get_logger()


async def run_prompt_to_bot_job(
    app: Any,
    job: JobInfo,
    prompt: str,
) -> None:
    """Run the prompt-to-bot job in the background.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        prompt: The natural language prompt for bot generation.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator

    await push_job_status_event(job, JobStatus.received)

    try:
        # 1. Generating
        await push_job_status_event(job, JobStatus.generating)
        commit_sha = await project_generator.generate_project_with_retries(
            prompt,
            template=ProjectTemplateName.BASIC,
        )
        bot_files = project_generator.get_bot_files()
        await push_job_status_event(job, JobStatus.generation_success)

        # 2. Training
        await push_job_status_event(job, JobStatus.training)
        agent = await train_and_load_and_link_agent(project_generator, commit_sha)
        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        # 3. Create copilot welcome message job
        copilot_welcome_job = job_manager.create_job(commit_sha)
        app.add_task(run_copilot_welcome_message_job(app, copilot_welcome_job))

        structlogger.info(
            "bot_builder_service.prompt_to_bot.success",
            files_generated=list(bot_files.keys()),
            copilot_welcome_job_id=copilot_welcome_job.id,
        )
        await push_job_status_event(
            job=job,
            status=JobStatus.done,
            payload={"copilot_welcome_job_id": copilot_welcome_job.id},
        )
        job_manager.mark_done(job)

    except TrainingError as exc:
        structlogger.debug(
            "prompt_to_bot_job.training_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.train_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except ValidationError as exc:
        # Log levels to include in the error message
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")

        structlogger.debug(
            "prompt_to_bot_job.validation_error",
            job_id=job.id,
            error=str(exc),
            all_validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )

        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        await push_job_status_event(
            job, JobStatus.validation_error, message=error_message
        )
        job_manager.mark_done(job, error=error_message)

    except (ProjectGenerationError, LLMGenerationError) as exc:
        structlogger.debug(
            "prompt_to_bot_job.generation_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.generation_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except Exception as exc:
        # Capture full traceback
        structlogger.exception(
            "prompt_to_bot_job.unexpected_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_template_to_bot_job(
    app: "Sanic",
    job: JobInfo,
    template_name: ProjectTemplateName,
) -> None:
    """Run the template-to-bot job in the background.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        template_name: The name of the template to use for bot generation.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator

    copilot_template_prompt_job = job_manager.create_job()
    await push_job_status_event(
        job,
        JobStatus.received,
        payload={"copilot_template_prompt_job_id": copilot_template_prompt_job.id},
    )

    try:
        app.add_task(
            run_copilot_template_prompt_job(
                app, copilot_template_prompt_job, template_name
            )
        )

        await push_job_status_event(job, JobStatus.generating)
        commit_sha = await project_generator.init_from_template(template_name)
        bot_files = project_generator.get_bot_files()
        await push_job_status_event(job, JobStatus.generation_success)

        await push_job_status_event(job, JobStatus.training)
        agent = await try_load_existing_agent(project_generator.project_folder)
        if agent is None:
            agent = await train_and_load_and_link_agent(project_generator, commit_sha)
        else:
            structlogger.info(
                "bot_builder_service.template_to_bot.agent_loaded_from_cache",
            )
        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        copilot_welcome_job = job_manager.create_job(commit_sha)
        app.add_task(
            run_copilot_welcome_message_job(app, copilot_welcome_job, template_name)
        )

        structlogger.info(
            "bot_builder_service.template_to_bot.success",
            files_generated=list(bot_files.keys()),
            copilot_template_prompt_job_id=copilot_template_prompt_job.id,
            copilot_welcome_job_id=copilot_welcome_job.id,
        )
        await push_job_status_event(
            job=job,
            status=JobStatus.done,
            payload={"copilot_welcome_job_id": copilot_welcome_job.id},
        )
        job_manager.mark_done(job)

    except TrainingError as exc:
        structlogger.debug(
            "template_to_bot_job.training_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.train_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except ValidationError as exc:
        # Log levels to include in the error message
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")

        structlogger.debug(
            "template_to_bot_job.validation_error",
            job_id=job.id,
            error=str(exc),
            all_validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )
        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        await push_job_status_event(
            job, JobStatus.validation_error, message=error_message
        )
        job_manager.mark_done(job, error=error_message)

    except ProjectGenerationError as exc:
        structlogger.debug(
            "template_to_bot_job.generation_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.generation_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except Exception as exc:
        # Capture full traceback
        structlogger.exception(
            "template_to_bot_job.unexpected_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_replace_all_files_job(
    app: "Sanic",
    job: JobInfo,
    bot_files: Dict[str, Any],
    commit_info: GitCommitInfo,
) -> None:
    """Run the replace-all-files job in the background.

    This replaces all bot files with the provided files and deletes any files
    not included in the request (excluding .rasa/ and models/ directories).

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        bot_files: Dictionary of file names to content for replacement.
        commit_info: Optional arguments for the commit
    """
    project_generator: ProjectGenerator = app.ctx.project_generator
    await push_job_status_event(job, JobStatus.received)

    commit_sha: Optional[str] = None
    try:
        commit_sha = await project_generator.replace_all_bot_files(
            bot_files, commit_info
        )
        # Validating
        await push_job_status_event(job, JobStatus.validating)
        training_input = project_generator.get_training_input()
        validation_error = await validate_project(training_input.importer)
        if validation_error:
            raise ValidationError(validation_error)
        await push_job_status_event(job, JobStatus.validation_success)

        # Training
        await push_job_status_event(job, JobStatus.training)
        agent = await train_and_load_and_link_agent(project_generator, commit_sha)
        update_agent(agent, app)

        await push_job_status_event(job, JobStatus.train_success)

        # Send final done event with copilot training success response job ID
        copilot_training_success_job = job_manager.create_job(commit_sha=commit_sha)
        app.add_task(
            run_copilot_training_success_job(app, copilot_training_success_job)
        )

        await push_job_status_event(
            job=job,
            status=JobStatus.done,
            payload={
                "copilot_training_success_job_id": copilot_training_success_job.id
            },
        )
        job_manager.mark_done(job)

    except GitOperationInProgressError as exc:
        error_message = str(exc)
        structlogger.debug(
            "replace_all_files_job.operation_in_progress",
            job_id=job.id,
            error=error_message,
        )
        await push_job_status_event(job, JobStatus.error, message=error_message)
        job_manager.mark_done(job, error=error_message)

    except ValidationError as exc:
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")
        structlogger.debug(
            "replace_all_files_job.validation_error",
            job_id=job.id,
            error=str(exc),
            validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )
        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        # Push error event and start copilot analysis job
        await push_error_and_start_copilot_analysis(
            app,
            job,
            JobStatus.validation_error,
            error_message,
            bot_files,
            commit_sha,
        )

        # After error mark job as done
        job_manager.mark_done(job, error=error_message)

    except TrainingError as exc:
        error_message = str(exc)
        structlogger.debug(
            "replace_all_files_job.train_error",
            job_id=job.id,
            error=error_message,
        )
        # Push error event and start copilot analysis job
        await push_error_and_start_copilot_analysis(
            app,
            job,
            JobStatus.train_error,
            error_message,
            bot_files,
            commit_sha,
        )

        # After error mark job as done
        job_manager.mark_done(job, error=error_message)

    except Exception as exc:
        # Capture full traceback for anything truly unexpected
        error_message = str(exc)
        structlogger.exception(
            "replace_all_files_job.unexpected_error",
            job_id=job.id,
            error=error_message,
        )

        # Push error event and start copilot analysis job
        await push_error_and_start_copilot_analysis(
            app,
            job,
            JobStatus.error,
            error_message,
            bot_files,
            commit_sha,
        )

        # After error mark job as done
        job_manager.mark_done(job, error=str(exc))


async def run_copilot_training_error_analysis_job(
    app: "Sanic",
    job: JobInfo,
    training_error_message: str,
    bot_files: Dict[str, Any],
) -> None:
    """Run copilot training error analysis job."""
    await push_job_status_event(job, JobStatus.received)

    try:
        # Create message content blocks with log content and available files
        log_content_block = LogContent(
            type="log", content=training_error_message, context="training_error"
        )
        file_content_blocks = [
            FileContent(type="file", file_path=file_path, file_content=file_content)
            for file_path, file_content in bot_files.items()
        ]
        context = CopilotContext(
            tracker_context=None,  # No conversation context needed
            copilot_chat_history=[
                InternalCopilotRequestChatMessage(
                    role="internal_copilot_request",
                    content=[log_content_block, *file_content_blocks],
                    response_category=ResponseCategory.TRAINING_ERROR_LOG_ANALYSIS,
                )
            ],
        )

        # Generate copilot response
        copilot_client = llm_service.instantiate_copilot()
        (
            original_stream,
            generation_context,
        ) = await copilot_client.generate_response(context)

        copilot_response_handler = llm_service.instantiate_handler(
            config.COPILOT_HANDLER_ROLLING_BUFFER_SIZE
        )
        intercepted_stream = copilot_response_handler.handle_response(original_stream)

        commit_info = None
        if job.commit_sha:
            # Get commit info
            commit_info = await app.ctx.project_generator.git_service.get_commit_info(
                job.commit_sha
            )

            commit_info["training_success"] = False

            # Send commit
            await push_job_status_event(
                job,
                JobStatus.copilot_analyzing,
                payload={
                    "response_category": "copilot",
                    "commit": commit_info,
                    "completeness": "complete",
                },
            )

        # Send original error log as part of copilot_analyzing stream
        training_error_log = TrainingErrorLog(logs=[log_content_block])
        await push_job_status_event(
            job, JobStatus.copilot_analyzing, payload=training_error_log.sse_data
        )

        # Stream the copilot response as job events
        async for token in intercepted_stream:
            # Send each token as a job event using the same format as /copilot endpoint
            await push_job_status_event(
                job, JobStatus.copilot_analyzing, payload=token.sse_data
            )

        # Send references (if any) as part of copilot_analyzing stream
        if generation_context.relevant_documents:
            reference_section = copilot_response_handler.extract_references(
                generation_context.relevant_documents
            )
            await push_job_status_event(
                job, JobStatus.copilot_analyzing, payload=reference_section.sse_data
            )

        # Persist the training error analysis to history
        full_text, _ = copilot_response_handler.extract_full_text_and_category()

        # Extract references if available
        references = None
        if generation_context.relevant_documents:
            reference_section = copilot_response_handler.extract_references(
                generation_context.relevant_documents
            )
            references = (
                reference_section.references if reference_section.references else None
            )

        await persist_training_error_analysis_to_history(
            text=full_text,
            logs=[log_content_block] if log_content_block else None,
            references=references,
            response_category=ResponseCategory.TRAINING_ERROR_LOG_ANALYSIS,
            commit=commit_info,
        )

        # Send success status
        await push_job_status_event(job, JobStatus.copilot_analysis_success)

        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

    except Exception as exc:
        structlogger.exception(
            "copilot_training_error_analysis_job.error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(
            job, JobStatus.copilot_analysis_error, message=str(exc)
        )
        job_manager.mark_done(job, error=str(exc))


async def run_copilot_welcome_message_job(
    app: "Sanic",
    job: JobInfo,
    template_name: Optional[ProjectTemplateName] = None,
) -> None:
    """Run the welcome message job in the background.

    This job sends a welcome message to the user after successful bot creation.
    For template-based bots, it sends a predefined message.
    For prompt-based bots, it can be extended to stream generated messages.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        template_name: The template name for template-based bots, None for prompt-based.
    """
    try:
        # Load welcome messages from YAML
        welcome_messages = load_copilot_welcome_messages()

        # Get the appropriate welcome message
        if template_name:
            welcome_message = welcome_messages.get(
                template_name.value,
                welcome_messages.get(PROMPT_TO_BOT_KEY),
            )
        else:
            welcome_message = welcome_messages.get(PROMPT_TO_BOT_KEY)

        commit_info = None
        if job.commit_sha:
            # Get commit info
            commit_info = await app.ctx.project_generator.git_service.get_commit_info(
                job.commit_sha
            )

            commit_info["training_success"] = True

            # Send commit
            await push_job_status_event(
                job,
                JobStatus.copilot_welcome_message,
                payload={
                    "response_category": "copilot",
                    "commit": commit_info,
                    "completeness": "complete",
                },
            )

        # Send the welcome message as a single event
        await push_job_status_event(
            job,
            JobStatus.copilot_welcome_message,
            payload={
                "content": welcome_message,
                "response_category": "copilot",
                "completeness": "complete",
            },
        )

        # Persist the welcome message to conversation history
        await persist_copilot_message_to_history(
            text=welcome_message, commit=commit_info
        )

        # Mark job as done
        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

        structlogger.info(
            "copilot_welcome_message_job.success",
            job_id=job.id,
            template=template_name.value if template_name else PROMPT_TO_BOT_KEY,
        )

    except Exception as exc:
        structlogger.exception(
            "welcome_message_job.error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_copilot_training_success_job(
    app: "Sanic",
    job: JobInfo,
) -> None:
    """Run the training success job in the background.

    This job sends a training success message to the user after successful bot training.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
    """
    try:
        # Load copilot default messages from YAML
        internal_messages = load_copilot_handler_default_responses()

        # Get the appropriate training success message
        training_success_message = internal_messages.get("training_success_response")

        commit_info = None
        if job.commit_sha:
            # Get commit info
            commit_info = await app.ctx.project_generator.git_service.get_commit_info(
                job.commit_sha
            )

            # Add training success
            commit_info["training_success"] = True

            # Send the training success message
            await push_job_status_event(
                job,
                JobStatus.train_success_message,
                payload={
                    "response_category": "copilot",
                    "commit": commit_info,
                    "completeness": "complete",
                },
            )

        # Send the training success message
        await push_job_status_event(
            job,
            JobStatus.train_success_message,
            payload={
                "content": training_success_message,
                "response_category": "copilot",
                "completeness": "complete",
            },
        )

        await persist_copilot_message_to_history(
            text=training_success_message, commit=commit_info
        )

        # Mark job as done
        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

        structlogger.info(
            "copilot_training_success_job.success",
            job_id=job.id,
        )

    except Exception as exc:
        structlogger.exception(
            "copilot_training_success_job.error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_copilot_go_back_in_time_success_job(
    app: "Sanic",
    job: JobInfo,
    internal_message_key: str,
    status: JobStatus,
) -> None:
    """Run the rollback success job in the background.

    This job sends a rollback success message to the user after successful rollback.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        internal_message_key: The key of the internal message to send.
        status: The status of the job.

    Returns:
        None
    """
    try:
        # Load copilot default messages from YAML
        internal_messages = load_copilot_handler_default_responses()

        # Get the appropriate rollback success message
        success_message = internal_messages.get(internal_message_key)

        commit_info = None
        if job.commit_sha:
            # Get commit info
            commit_info = await app.ctx.project_generator.git_service.get_commit_info(
                job.commit_sha
            )

            commit_info["training_success"] = True

            # Send commit
            await push_job_status_event(
                job,
                status,
                payload={
                    "response_category": "copilot",
                    "commit": commit_info,
                    "completeness": "complete",
                },
            )

        # Send the success message
        await push_job_status_event(
            job,
            status,
            payload={
                "content": success_message,
                "response_category": "copilot",
                "completeness": "complete",
            },
        )

        await persist_copilot_message_to_history(
            text=success_message, commit=commit_info
        )

        # Mark job as done
        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

        structlogger.info(
            "copilot_go_back_in_time_success_job.success",
            job_id=job.id,
        )

    except Exception as exc:
        structlogger.exception(
            "copilot_go_back_in_time_success_job.error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_change_branch_job(
    app: "Sanic",
    job: JobInfo,
    branch_name: str,
    create_if_not_exists: bool = False,
) -> None:
    """Run the change branch job in the background.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        branch_name: The branch name to checkout.
        create_if_not_exists: Whether to create the branch if it doesn't exist.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator

    await push_job_status_event(job, JobStatus.received)

    try:
        # 1. Switch branch
        await push_job_status_event(job, JobStatus.switching_branch)
        commit_sha = await project_generator.checkout_branch(
            branch_name, create_if_not_exists
        )
        await push_job_status_event(job, JobStatus.branch_switch_success)

        # 2. Training
        await push_job_status_event(job, JobStatus.training)
        agent = await load_or_train_agent_for_commit(project_generator, job, commit_sha)
        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        # 3. Done
        structlogger.info(
            "bot_builder_service.change_branch.success",
            branch_name=branch_name,
        )
        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

    except GitOperationInProgressError as exc:
        structlogger.debug(
            "change_branch_job.operation_in_progress",
            job_id=job.id,
            error=str(exc),
            branch_name=branch_name,
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except subprocess.CalledProcessError as exc:
        structlogger.debug(
            "change_branch_job.branch_switch_error",
            job_id=job.id,
            error=str(exc),
            branch_name=branch_name,
        )
        await push_job_status_event(
            job, JobStatus.branch_switch_error, message=str(exc)
        )
        job_manager.mark_done(job, error=str(exc))

    except TrainingError as exc:
        structlogger.debug(
            "change_branch_job.training_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.train_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except ValidationError as exc:
        # Log levels to include in the error message
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")

        structlogger.debug(
            "change_branch_job.validation_error",
            job_id=job.id,
            error=str(exc),
            all_validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )
        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        await push_job_status_event(
            job, JobStatus.validation_error, message=error_message
        )
        job_manager.mark_done(job, error=error_message)

    except Exception as exc:
        # Capture full traceback
        structlogger.exception(
            "change_branch_job.unexpected_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_rollback_job(
    app: "Sanic",
    job: JobInfo,
    commit_sha: str,
) -> None:
    """Run rollback job following existing patterns.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        commit_sha: SHA of the commit to rollback to.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator
    await push_job_status_event(job, JobStatus.received)

    try:
        rollback_commit_sha = await perform_rollback(project_generator, job, commit_sha)
        # we use the prior commit sha, since that saves us from a retrain in
        # case we have a trained model for the prior commit. using the
        # rollback sha wouldn't make sense as that is a new commit, so there
        # would surely no model be trained for that
        agent = await load_or_train_agent_for_commit(project_generator, job, commit_sha)

        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        structlogger.info(
            "bot_builder_service.rollback.success", commit_sha=rollback_commit_sha
        )
        copilot_rollback_success_job = job_manager.create_job(
            commit_sha=rollback_commit_sha
        )
        app.add_task(
            run_copilot_go_back_in_time_success_job(
                app,
                copilot_rollback_success_job,
                "rollback_success_response",
                JobStatus.rollback_success,
            )
        )

        await push_job_status_event(
            job,
            JobStatus.done,
            payload={
                "copilot_rollback_success_job_id": copilot_rollback_success_job.id
            },
        )
        job_manager.mark_done(job)

    except GitOperationInProgressError as exc:
        await handle_rollback_error(job, exc, commit_sha, JobStatus.error)
    except subprocess.CalledProcessError as exc:
        await handle_rollback_error(job, exc, commit_sha, JobStatus.rollback_error)
    except TrainingError as exc:
        await handle_rollback_error(job, exc, commit_sha, JobStatus.train_error)
    except Exception as exc:
        await handle_rollback_error(
            job, exc, commit_sha, JobStatus.error, log_traceback=True
        )


async def run_revert_job(
    app: "Sanic",
    job: JobInfo,
    commit_sha: str,
) -> None:
    """Run revert job following existing patterns.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        commit_sha: SHA of the commit to revert.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator
    await push_job_status_event(job, JobStatus.received)

    try:
        revert_commit_sha = await perform_revert(project_generator, job, commit_sha)
        # we use the prior commit sha, since that saves us from a retrain in
        # case we have a trained model for the prior commit. using the
        # revert sha wouldn't make sense as that is a new commit, so there
        # would surely no model be trained for that
        agent = await load_or_train_agent_for_commit(project_generator, job, commit_sha)

        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        structlogger.info("bot_builder_service.revert.success", commit_sha=commit_sha)
        copilot_revert_success_job = job_manager.create_job(
            commit_sha=revert_commit_sha
        )
        app.add_task(
            run_copilot_go_back_in_time_success_job(
                app,
                copilot_revert_success_job,
                "revert_success_response",
                JobStatus.revert_success,
            )
        )

        await push_job_status_event(
            job=job,
            status=JobStatus.done,
            payload={"copilot_revert_success_job_id": copilot_revert_success_job.id},
        )
        job_manager.mark_done(job)

    except GitOperationInProgressError as exc:
        await handle_revert_error(job, exc, commit_sha, JobStatus.error)
    except subprocess.CalledProcessError as exc:
        await handle_revert_error(job, exc, commit_sha, JobStatus.revert_error)
    except TrainingError as exc:
        await handle_revert_error(job, exc, commit_sha, JobStatus.train_error)
    except Exception as exc:
        await handle_revert_error(
            job, exc, commit_sha, JobStatus.error, log_traceback=True
        )


async def run_copilot_template_prompt_job(
    app: "Sanic",
    job: JobInfo,
    template_name: ProjectTemplateName,
) -> None:
    """Run the template prompt job in the background.

    This job sends the template prompt as a user message at the start
    of a template-based bot creation.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        template_name: The template name to get the prompt for.
    """
    try:
        template_prompts = load_copilot_template_prompts()
        template_prompt = template_prompts.get(template_name.value)

        if not template_prompt:
            structlogger.warning(
                "copilot_template_prompt_job.no_prompt_found",
                job_id=job.id,
                template=template_name.value,
            )
            await push_job_status_event(job, JobStatus.done)
            job_manager.mark_done(job)
            return

        await push_job_status_event(
            job,
            JobStatus.copilot_template_prompt,
            payload={
                "content": template_prompt,
                "completeness": "complete",
            },
        )

        await persist_user_message_to_history(text=template_prompt)

        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

        structlogger.info(
            "copilot_template_prompt_job.success",
            job_id=job.id,
            template=template_name.value,
        )

    except Exception as exc:
        structlogger.exception(
            "copilot_template_prompt_job.error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


def _safe_tar_members(
    tar: tarfile.TarFile, destination_directory: Path
) -> List[tarfile.TarInfo]:
    """Get safe members for extraction to prevent path traversal and resource attacks.

    Args:
        tar: Open tar file handle
        destination_directory: Directory to which files will be extracted

    Returns:
        List of members that are safe to extract within destination_directory

    Raises:
        ProjectGenerationError: If archive violates security constraints
    """
    base_path = destination_directory.resolve()
    safe_members = []
    total_size = 0
    file_count = 0

    for member in tar.getmembers():
        name = member.name

        # Check file count limit
        file_count += 1
        if file_count > MAX_ARCHIVE_FILES:
            raise ProjectGenerationError(
                f"Archive contains too many files (>{MAX_ARCHIVE_FILES}).", attempts=1
            )

        # Skip empty names and absolute paths
        if not name or name.startswith("/") or name.startswith("\\"):
            continue

        # Disallow symlinks and hardlinks
        if member.issym() or member.islnk():
            continue

        # Check individual file size limit
        if member.size > MAX_ARCHIVE_FILE_SIZE:
            raise ProjectGenerationError(
                f"Archive contains file '{name}' that is too large "
                f"({member.size} bytes > {MAX_ARCHIVE_FILE_SIZE} bytes).",
                attempts=1,
            )

        # Check total size limit
        total_size += member.size
        if total_size > MAX_ARCHIVE_TOTAL_SIZE:
            raise ProjectGenerationError(
                "Archive total size too large "
                f"({total_size} bytes > {MAX_ARCHIVE_TOTAL_SIZE} bytes).",
                attempts=1,
            )

        # Compute the final path and ensure it's within base_path
        target_path = (base_path / name).resolve()
        try:
            target_path.relative_to(base_path)
        except ValueError:
            # Member would escape the destination directory
            continue

        safe_members.append(member)

    return safe_members


async def run_backup_to_bot_job(
    app: "Sanic",
    job: JobInfo,
    presigned_url: str,
) -> None:
    """Run the backup-to-bot job in the background.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        presigned_url: Presigned URL to download tar.gz backup data.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator
    await push_job_status_event(job, JobStatus.received)

    temp_file_path = None
    try:
        # 1) Download and extract backup
        await push_job_status_event(job, JobStatus.generating)
        temp_file_path = await download_backup_from_url(presigned_url)

        # Clear existing project files, keeping .rasa and __pycache__
        project_path = Path(project_generator.project_folder)
        project_generator.cleanup(skip_files=[".rasa", "__pycache__"])

        # Extract the backup archive
        with tarfile.open(temp_file_path, "r:gz") as tar:
            safe_members = _safe_tar_members(tar, project_path)
            tar.extractall(path=project_path, members=safe_members)  # nosec B202:tarfile_unsafe_members

        await push_job_status_event(job, JobStatus.generation_success)

        # Ensure Git is initialized
        project_generator.migrate_git_repository_if_needed()
        commit_sha = await project_generator.git_service.get_current_commit_sha()
        agent = await load_or_train_agent_for_commit(project_generator, job, commit_sha)

        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        structlogger.info(
            "bot_builder_service.backup_to_bot.success",
            commit_sha=commit_sha,
        )
        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

    except tarfile.ReadError as exc:
        raise ProjectGenerationError(
            f"Failed to extract backup archive: {exc}. "
            f"Please ensure the backup file is a valid tar.gz archive.",
            attempts=1,
        )
    except TrainingError as exc:
        structlogger.debug(
            "backup_to_bot_job.training_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.train_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except ValidationError as exc:
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")

        structlogger.debug(
            "backup_to_bot_job.validation_error",
            job_id=job.id,
            error=str(exc),
            all_validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )
        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        await push_job_status_event(
            job, JobStatus.validation_error, message=error_message
        )
        job_manager.mark_done(job, error=error_message)

    except ProjectGenerationError as exc:
        structlogger.debug(
            "backup_to_bot_job.generation_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.generation_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except Exception as exc:
        structlogger.exception(
            "backup_to_bot_job.unexpected_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))
    finally:
        # Always clean up temp file
        if temp_file_path:
            try:
                Path(temp_file_path).unlink(missing_ok=True)
            except Exception:
                pass



import logging
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Text, Union, cast

from rasa.core.secrets_manager.constants import (
    SECRET_KEY_LABEL,
    SECRET_MANAGER_PREFIX,
    SOURCE_KEY_LABEL,
    SUPPORTED_SECRET_MANAGERS,
)
from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)


class CredentialsLocation:
    """Points to where secret is stored.

    This is a DTO object which is used to pass the configuration to the secret manager.
    More specific configuration is created based on this DTO and
    a secret manager which is used.
    For example: VaultSecretsManager will create VaultStoredSecret
    based on this DTO.
    """

    def __init__(self, source: Text, secret_key: Text, **kwargs: Optional[Any]) -> None:
        """Initialise the CredentialsLocation.

        Args:
            source: The source of the secret.
            secret_key: The key of the secret.
            **kwargs: Additional arguments.
        """
        self.source = source
        self.secret_key = secret_key
        self.kwargs = kwargs

    def get_secret_manager_name(self) -> Text:
        """Get the name of the targeted secret manager.

        Returns:
            The name of the secret manager.
        """
        return self.source.replace(f"{SECRET_MANAGER_PREFIX}.", "")

    @staticmethod
    def is_credentials_location_instance(
        value: Union[Text, "CredentialsLocation"],
    ) -> bool:
        """Check if the value is a CredentialsLocation.

        Args:
            value: The value to check.

        Returns:
            True if the value is an instance  CredentialsLocation.
        """
        return isinstance(value, CredentialsLocation)

    @staticmethod
    def is_property_non_empty_string(
        credentials_location: Dict[Text, Text], property_name: Text
    ) -> bool:
        """Check if the property of credentials location is valid.

        Args:
            credentials_location: The credentials location to check.
            property_name: The property to check.

        Returns:
            True if the property is valid.
        """
        if property_name not in credentials_location.keys():
            return False

        if not isinstance(credentials_location.get(property_name), str):
            return False

        if credentials_location.get(property_name) == "":
            return False

        return True

    @staticmethod
    def is_credentials_location_valid(
        raw_credentials_location: Dict[Text, Text],
    ) -> bool:
        """Check if the configuration is a secret manager configuration.

        Args:
            raw_credentials_location: A dictionary of values to check.

        Returns:
            True if the value is a valid credentials' location.
        """
        if not isinstance(raw_credentials_location, dict):
            return False

        if not CredentialsLocation._is_source_valid(
            raw_credentials_location=raw_credentials_location
        ):
            return False

        if not CredentialsLocation.is_property_non_empty_string(
            credentials_location=raw_credentials_location,
            property_name=SECRET_KEY_LABEL,
        ):
            return False

        return True

    @staticmethod
    def _is_source_valid(raw_credentials_location: Dict[Text, Text]) -> bool:
        """Check if the source is valid.

        Args:
            raw_credentials_location: A dictionary of values to check.

        Returns:
            True if the source is valid.
        """
        is_source_non_empty_string = CredentialsLocation.is_property_non_empty_string(
            credentials_location=raw_credentials_location,
            property_name=SOURCE_KEY_LABEL,
        )

        if not is_source_non_empty_string:
            return False

        source_value = raw_credentials_location.get(SOURCE_KEY_LABEL)
        if not source_value or source_value == "":
            return False

        parts = source_value.split(".")

        return (
            is_source_non_empty_string
            and len(parts) == 2
            and parts[0] == SECRET_MANAGER_PREFIX
            and parts[1] != ""
        )

    @staticmethod
    def are_required_properties_valid(
        value: Dict[Text, Any], endpoint_property: Text
    ) -> bool:
        """Check if required properties of raw credentials location are valid.

        Args:
            value: a raw credentials location value
            endpoint_property: name of the property to which raw
                credentials value is assigned to

        Returns:
            True if source and secret_key are valid.
            Otherwise, it returns False.
        """
        if not CredentialsLocation.is_property_non_empty_string(
            value, SOURCE_KEY_LABEL
        ):
            logger.error(
                f"Property '{SOURCE_KEY_LABEL}' is missing or invalid for "
                f"'{endpoint_property}'."
            )
            return False

        if not CredentialsLocation.is_property_non_empty_string(
            value, SECRET_KEY_LABEL
        ):
            logger.error(
                f"Property '{SECRET_KEY_LABEL}' is missing or invalid for "
                f"'{endpoint_property}'."
            )
            return False

        return True


@dataclass
class EndpointTrait:
    """Represents a trait of an endpoint.

    Args:
        endpoint_config: The config for an endpoint.
        endpoint_type: The type of the endpoint. Like tracker_store, event_broker etc.
    """

    endpoint_config: "EndpointConfig"
    endpoint_type: Text


class EndpointReader:
    """Reads and validates parts of endpoint configuration.

    This is a helper class for the `EndpointConfig` class.
    """

    def __init__(self, endpoint_config: "EndpointConfig"):
        """Initialise the endpoint reader.

        Args:
            endpoint_config: The configuration for an endpoint.
        """
        self.endpoint_config = endpoint_config

    def get_property_value(
        self,
        endpoint_property_name: Text,
    ) -> Optional[Union[str, CredentialsLocation]]:
        """Return a value stored in the endpoint's property.

        It is constrained to work only on string values and dictionaries which
        hold credentials' location.

        Args:
            endpoint_property_name: The endpoint's property name.

        Returns:
            The value of the property. It can be a string or a CredentialsLocation.
            We have constrained it to work only on string values and dictionaries
            because they are the only types which are supported by the
            endpoint's properties we are currently interested in.
        """
        endpoint_property_value = self._get_raw_property_value(endpoint_property_name)

        if (
            CredentialsLocation.is_credentials_location_valid(
                raw_credentials_location=endpoint_property_value
            )
            and endpoint_property_value.get(SOURCE_KEY_LABEL).replace(
                f"{SECRET_MANAGER_PREFIX}.", ""
            )
            in SUPPORTED_SECRET_MANAGERS
        ):
            return CredentialsLocation(
                **endpoint_property_value,
            )

        return endpoint_property_value

    def _get_raw_property_value(self, property_name: Text) -> Any:
        """Get the raw value of a property in EndpointsConfig.

        Args:
            property_name: The key of the property.

        Returns:
            The raw value of the property.
        """
        # check if it is stored in kwargs
        property_value = self.endpoint_config.kwargs.get(property_name)

        if property_value is None:
            # check if it is an attribute of EndpointsConfig
            if property_name in self.endpoint_config.__dict__.keys():
                property_value = self.endpoint_config.__dict__.get(property_name)
            else:
                raise RasaException(
                    f"Property {property_name} is not defined in the "
                    f"endpoint's configuration."
                )

        return property_value

    def has_property(
        self,
        property_name: Text,
    ) -> bool:
        """Return True if an endpoint has a given property.

        Otherwise, returns False.
        """
        return property_name in self.get_keys()

    def is_endpoint_property_valid(
        self,
        endpoint_property: Text,
    ) -> bool:
        """Validate the value of endpoint's property.

        Args:
            endpoint_property: name of the property to validate

        Returns:
            True if the value of the property is valid.
            Otherwise, it returns False.
        """
        value = self.get_property_value(endpoint_property_name=endpoint_property)
        if isinstance(value, str) and value == "":
            logger.error(f"Property '{endpoint_property}' is empty.")
            return False

        if isinstance(value, dict):
            return CredentialsLocation.are_required_properties_valid(
                value=value, endpoint_property=endpoint_property
            )

        return True

    def get_endpoint_properties_managed_by_secret_manager(
        self,
    ) -> Optional[Dict[Text, List[Text]]]:
        """Get endpoint's properties that are managed by the secret manager.

        Returns:
            A dictionary of the endpoint's properties that are managed
            by the secret manager.
            The key is the name of the secret manager and the value is a list of the
            endpoint's properties that are managed by the secret manager.

            For example, if the endpoint's properties `username` and `password`
            are managed by the Vault secret manager, the returned dictionary will be:
            {
                vault: ["username", "password"]
            }
        """
        config_keys: Dict[Text, List[Text]] = {}

        for config_key_name in self.get_keys():
            config = self.get_property_value(endpoint_property_name=config_key_name)

            if config and CredentialsLocation.is_credentials_location_instance(config):
                credentials_location = cast(CredentialsLocation, config)
                secret_manager_name = credentials_location.source.replace(
                    f"{SECRET_MANAGER_PREFIX}.", ""
                )
                if secret_manager_name not in config_keys.keys():
                    config_keys[secret_manager_name] = []

                config_keys[secret_manager_name].append(config_key_name)

        return config_keys if len(config_keys.keys()) > 0 else None

    def get_keys(self) -> List[Text]:
        """Get the list of all properties of the endpoint's configuration.

        Returns:
            The list of all properties of the endpoint's configuration.
        """
        return [key for key in self.endpoint_config.kwargs.keys()] + [
            key for key in self.endpoint_config.__dict__.keys()
        ]


class TrackerStoreEndpointValidator:
    """Processes the tracker store key's related to vault configuration."""

    USERNAME_PROPERTY = "username"
    PASSWORD_PROPERTY = "password"

    def __init__(self, endpoint_reader: EndpointReader):
        """Initialise the TrackerStoreEndpointValidator.

        Args:
            endpoint_reader: The reader of endpoint's configuration.
        """
        self.endpoint_reader = endpoint_reader

    def is_endpoint_config_valid(self) -> bool:
        """Check if the endpoint config is valid.

        Returns:
            True if the endpoint config is valid, False otherwise.
        """
        return self._is_username_valid() and self._is_password_valid()

    def _is_username_valid(self) -> bool:
        """Check if the username property is valid.

        Returns:
            True if the username property is valid, False otherwise.
        """
        return not self.endpoint_reader.has_property(
            property_name=self.USERNAME_PROPERTY
        ) or self.endpoint_reader.is_endpoint_property_valid(
            endpoint_property=self.USERNAME_PROPERTY
        )

    def _is_password_valid(self) -> bool:
        """Check if the password property is valid.

        Returns:
            True if the password property is valid, False otherwise.
        """
        return self.endpoint_reader.has_property(
            property_name=self.PASSWORD_PROPERTY
        ) and self.endpoint_reader.is_endpoint_property_valid(
            endpoint_property=self.PASSWORD_PROPERTY
        )



from dataclasses import dataclass
from typing import Any, Dict, Text

from packaging import version

from rasa.constants import MINIMUM_COMPATIBLE_VERSION
from rasa.shared.exceptions import RasaException


@dataclass
class UnsupportedModelVersionError(RasaException):
    """Raised when a model is too old to be loaded.

    Args:
        model_version: the used model version that is not supported and triggered
            this exception
    """

    model_version: Text

    def __str__(self) -> Text:
        minimum_version = version.parse(MINIMUM_COMPATIBLE_VERSION)
        return (
            f"The model version is trained using Rasa Pro {self.model_version} "
            f"and is not compatible with your current installation "
            f"which supports models build with Rasa Pro {minimum_version} "
            f"or higher. "
            f"This means that you either need to retrain your model "
            f"or revert back to the Rasa version that trained the model "
            f"to ensure that the versions match up again."
        )


class ModelNotFound(RasaException):
    """Raised when a model is not found in the path provided by the user."""


class NoEventsToMigrateError(RasaException):
    """Raised when no events to be migrated are found."""


class NoConversationsInTrackerStoreError(RasaException):
    """Raised when a tracker store does not contain any conversations."""


class NoEventsInTimeRangeError(RasaException):
    """Raised when a tracker store does not contain events within a given time range."""


class MissingDependencyException(RasaException):
    """Raised if a python package dependency is needed, but not installed."""


class DetailedRasaException(RasaException):
    """Base class for exceptions that carry an error code and extra context."""

    def __init__(self, *, code: str, event_info: str, **ctx: Any) -> None:
        super().__init__(event_info)
        self.code: str = code
        self.info: str = event_info
        self.ctx: Dict[str, Any] = ctx

    def __str__(self) -> str:
        return self.info


class HealthCheckError(DetailedRasaException):
    """Raised when an error occurs during health checks."""


class EnterpriseSearchPolicyError(DetailedRasaException):
    """Raised when an error occurs in EnterpriseSearchPolicy."""


class ValidationError(DetailedRasaException):
    """Raised when an error occurs during validation."""


@dataclass
class PublishingError(RasaException):
    """Raised when publishing of an event fails.

    Attributes:
        timestamp -- Unix timestamp of the event during which publishing fails.
    """

    timestamp: float

    def __str__(self) -> Text:
        """Returns string representation of exception."""
        return str(self.timestamp)


class ActionLimitReached(RasaException):
    """Raised when predicted action limit is reached."""



from typing import Text

from rasa.shared.exceptions import RasaCoreException, RasaException


class AgentNotReady(RasaCoreException):
    """Raised if someone tries to use an agent that is not ready.

    An agent might be created, e.g. without an processor attached. But
    if someone tries to parse a message with that agent, this exception
    will be thrown.
    """

    def __init__(self, message: Text) -> None:
        """Initialize message attribute."""
        self.message = message
        super(AgentNotReady, self).__init__(message)


class ChannelConfigError(RasaCoreException):
    """Raised if a channel is not configured correctly."""


class InvalidTrackerFeaturizerUsageError(RasaCoreException):
    """Raised if a tracker featurizer is incorrectly used."""


class KafkaProducerInitializationError(RasaException):
    """Raised if the Kafka Producer cannot be properly initialized."""



from typing import Dict, List, Set, Type

from rasa.dialogue_understanding.commands import (
    CancelFlowCommand,
    CannotHandleCommand,
    ChitChatAnswerCommand,
    Command,
    CorrectSlotsCommand,
    HumanHandoffCommand,
    KnowledgeAnswerCommand,
    RestartCommand,
    SessionStartCommand,
    SetSlotCommand,
    SkipQuestionCommand,
)
from rasa.dialogue_understanding.commands.user_silence_command import UserSilenceCommand
from rasa.dialogue_understanding.patterns.cancel import CancelPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.cannot_handle import (
    CannotHandlePatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.chitchat import ChitchatPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.human_handoff import (
    HumanHandoffPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.restart import RestartPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.search import SearchPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.session_start import (
    SessionStartPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.skip_question import (
    SkipQuestionPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.user_silence import (
    UserSilencePatternFlowStackFrame,
)

triggerable_pattern_to_command_class: Dict[str, Type[Command]] = {
    SessionStartPatternFlowStackFrame.flow_id: SessionStartCommand,
    UserSilencePatternFlowStackFrame.flow_id: UserSilenceCommand,
    CancelPatternFlowStackFrame.flow_id: CancelFlowCommand,
    ChitchatPatternFlowStackFrame.flow_id: ChitChatAnswerCommand,
    HumanHandoffPatternFlowStackFrame.flow_id: HumanHandoffCommand,
    SearchPatternFlowStackFrame.flow_id: KnowledgeAnswerCommand,
    SkipQuestionPatternFlowStackFrame.flow_id: SkipQuestionCommand,
    CannotHandlePatternFlowStackFrame.flow_id: CannotHandleCommand,
    RestartPatternFlowStackFrame.flow_id: RestartCommand,
}


def filter_slot_commands(
    commands: List[Command], overlapping_slot_names: Set[str]
) -> List[Command]:
    """Filter out slot commands that set overlapping slots."""
    filtered_commands = []

    for command in commands:
        if (
            isinstance(command, SetSlotCommand)
            and command.name in overlapping_slot_names
        ):
            continue

        if isinstance(command, CorrectSlotsCommand):
            allowed_slots = [
                slot
                for slot in command.corrected_slots
                if slot.name not in overlapping_slot_names
            ]
            if not allowed_slots:
                continue

            command.corrected_slots = allowed_slots

        filtered_commands.append(command)

    return filtered_commands
